@@ -3,8 +3,8 @@
  *   Copyright (c) 1998-2001 Michal Trojnara <Michal.Trojnara@mirt.net>
  *                 All Rights Reserved
  *
- *   Version:      3.20                  (stunnel.c)
- *   Date:         2001.08.15
+ *   Version:      3.22             (stunnel.c)
+ *   Date:         2001.12.20
  *
  *   Author:       Michal Trojnara  <Michal.Trojnara@mirt.net>
  *
@@ -21,20 +21,27 @@
  *   You should have received a copy of the GNU General Public License
  *   along with this program; if not, write to the Free Software
  *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ *
+ *   In addition, as a special exception, Michal Trojnara gives
+ *   permission to link the code of this program with the OpenSSL
+ *   library (or with modified versions of OpenSSL that use the same
+ *   license as OpenSSL), and distribute linked combinations including
+ *   the two.  You must obey the GNU General Public License in all
+ *   respects for all of the code used other than OpenSSL.  If you modify
+ *   this file, you may extend this exception to your version of the
+ *   file, but you are not obligated to do so.  If you do not wish to
+ *   do so, delete this exception statement from your version.
  */
 
-/* Undefine if you have problems with make_sockets() */
-#define INET_SOCKET_PAIR
-
-/* Max number of children is limited by FD_SETSIZE */
-/* Do not increase it over 500 */
-#ifdef FD_SETSIZE
-#define MAX_CLIENTS    ((FD_SETSIZE-24)/2)
-#else
-#define MAX_CLIENTS    500
-#endif
-
 #include "common.h"
+#include "prototypes.h"
+#include "client.h"
+
+#ifdef HAVE_OPENSSL
+#include <openssl/crypto.h> /* for SSLeay_version */
+#else
+#include <crypto.h>
+#endif
 
 #ifdef USE_WIN32
 static struct WSAData wsa_state;
@@ -50,32 +57,46 @@ static void delete_pid();
 
     /* Socket functions */
 static int listen_local();
-#ifndef USE_WIN32
-static int make_sockets(int [2]);
-#endif
 
     /* Error/exceptions handling functions */
-static void ioerror(char *);
+void ioerror(char *);
 void sockerror(char *);
+void log_error(int, int, char *);
+static char *my_strerror(int);
 #ifdef USE_FORK
 static void sigchld_handler(int);
 #endif
 #ifndef USE_WIN32
-static void local_handler(int);
+void local_handler(int);
 static void signal_handler(int);
+#else
+static BOOL CtrlHandler(DWORD);
 #endif
 
 server_options options;
 
+#ifdef USE_WIN32
+/*
+int WINAPI WinMain (HINSTANCE hInstance, HINSTANCE hPrevInstance,
+    PSTR szCmdLine, int iCmdShow) {
+    return 0;
+}
+*/
+#endif
+
     /* Functions */
 int main(int argc, char* argv[]) { /* execution begins here 8-) */
     struct stat st; /* buffer for stat */
 
 #ifdef USE_WIN32
-    if(WSAStartup(0x0101, &wsa_state)!=0) {
+    if(WSAStartup(0x0101, &wsa_state)) {
         sockerror("WSAStartup");
         exit(1);
     }
+    if(!SetConsoleCtrlHandler((PHANDLER_ROUTINE)CtrlHandler, TRUE)) {
+        ioerror("SetConsoleCtrlHandler");
+        exit(1);
+    }
 #else
     signal(SIGPIPE, SIG_IGN); /* avoid 'broken pipe' signal */
     signal(SIGTERM, signal_handler);
@@ -115,28 +136,23 @@ int main(int argc, char* argv[]) { /* execution begins here 8-) */
     /* check if started from inetd */
     context_init(); /* initialize global SSL context */
     sthreads_init(); /* initialize threads */
-    log(LOG_NOTICE, "%s", STUNNEL_INFO);
-    if (options.option & OPT_DAEMON) {
-        /* client or server, daemon mode */
+    log(LOG_NOTICE, "%s", stunnel_info());
+    if(options.option & OPT_DAEMON) { /* daemon mode */
 #ifndef USE_WIN32
-        if (!(options.option & OPT_FOREGROUND))
+        if(!(options.option & OPT_FOREGROUND))
             daemonize();
         create_pid();
 #endif
         daemon_loop();
-    } else if ((options.option & OPT_CLIENT) &&
-        (options.option & OPT_PROGRAM)) {
-        /* client, program mode */
-        int local;
-        u32 ip = 0; /* local program or stdin/stdout */
-        if ((local = connect_local(ip)) >= 0) {
-            options.clients = 1;
-            client(local);
+    } else { /* inetd mode */
+        max_fds=16; /* Just in case */
+        d=calloc(max_fds, sizeof(FD));
+        if(!d) {
+            log(LOG_ERR, "Memory allocation failed");
+            exit(1);
         }
-    } else {
-        /* client or server, inetd mode */
         options.clients = 1;
-        client(STDIO_FILENO); /* rd fd=0, wr fd=1 */
+        client((void *)STDIO_FILENO); /* rd fd=0, wr fd=1 */
     }
     /* close SSL */
     context_free(); /* free global SSL context */
@@ -148,18 +164,43 @@ static void daemon_loop() {
     int ls, s;
     struct sockaddr_in addr;
     int addrlen;
+    int max_clients, fds_ulimit=-1;
 
+#if defined HAVE_SYSCONF
+    fds_ulimit=sysconf(_SC_OPEN_MAX);
+    if(fds_ulimit<0)
+        ioerror("sysconf");
+#elif defined HAVE_GETRLIMIT
+    struct rlimit rlim;
+    if(getrlimit(RLIMIT_NOFILE, &rlim)<0)
+        ioerror("getrlimit");
+    else
+        fds_ulimit=rlim.rlim_cur;
+    if(fds_ulimit==RLIM_INFINITY)
+        fds_ulimit=-1;
+#endif
+    if(fds_ulimit>=16 && fds_ulimit<FD_SETSIZE)
+        max_fds=fds_ulimit;
+    else
+        max_fds=FD_SETSIZE;
+    d=calloc(max_fds, sizeof(FD));
+    if(!d) {
+        log(LOG_ERR, "Memory allocation failed");
+        exit(1);
+    }
+    max_clients=max_fds>=256 ? max_fds*125/256 : (max_fds-6)/2;
+    log(LOG_NOTICE, "FD_SETSIZE=%d, file ulimit=%d%s -> %d clients allowed",
+        FD_SETSIZE, fds_ulimit, fds_ulimit<0?" (unlimited)":"", max_clients);
     ls=listen_local();
     options.clients=0;
-#ifndef USE_WIN32
 #ifdef USE_FORK
-    /* Main process will receive signals about dead children */
+    /* Handle signals about dead children */
     signal(SIGCHLD, sigchld_handler);
-#else /* defined USE_FORK */
-    /* Main process will receive signals about dead children of it's threads */
-    signal(SIGCHLD, local_handler);
 #endif /* defined USE_FORK */
-#endif /* ndefined USE_WIN32 */
+#ifdef USE_PTHREAD
+    /* Handle signals about dead local processes */
+    signal(SIGCHLD, local_handler);
+#endif /* defined USE_PTHREAD */
     while(1) {
         addrlen=sizeof(addr);
         do {
@@ -169,29 +210,36 @@ static void daemon_loop() {
             sockerror("accept");
             continue;
         }
-        if(options.clients<MAX_CLIENTS) {
-            if(create_client(ls, s, client)) {
-                enter_critical_section(CRIT_NTOA); /* inet_ntoa is not mt-safe */
-                log(LOG_WARNING,
-                    "%s create_client failed - connection from %s:%d REJECTED",
-                    options.servname,
-                    inet_ntoa(addr.sin_addr),
-                    ntohs(addr.sin_port));
-                leave_critical_section(CRIT_NTOA);
-            } else {
-                enter_critical_section(CRIT_CLIENTS); /* for multi-cpu machines */
-                options.clients++;
-                leave_critical_section(CRIT_CLIENTS);
-            }
-        } else {
-            enter_critical_section(CRIT_NTOA); /* inet_ntoa is not mt-safe */
-            log(LOG_WARNING,
-                "%s has too many clients - connection from %s:%d REJECTED",
-                options.servname, inet_ntoa(addr.sin_addr), ntohs(addr.sin_port));
-            leave_critical_section(CRIT_NTOA);
+        enter_critical_section(CRIT_NTOA); /* inet_ntoa is not mt-safe */
+        log(LOG_DEBUG, "%s accepted FD=%d from %s:%d", options.servname, s,
+            inet_ntoa(addr.sin_addr), ntohs(addr.sin_port));
+        leave_critical_section(CRIT_NTOA);
+        if(options.clients>=max_clients) {
+            log(LOG_WARNING, "Connection rejected: too many clients (>=%d)",
+                max_clients);
             closesocket(s);
+            continue;
         }
+        if(s>=max_fds) {
+            log(LOG_ERR,
+                "Connection rejected: file descriptor out of range (>=%d)",
+                max_fds);
+            closesocket(s);
+            continue;
+        }
+#ifdef FD_CLOEXEC
+        fcntl(s, F_SETFD, FD_CLOEXEC); /* close socket in child execvp */
+#endif
+        if(create_client(ls, s, client)) {
+            log(LOG_ERR, "Connection rejected: create_client failed");
+            closesocket(s);
+            continue;
+        }
+        enter_critical_section(CRIT_CLIENTS); /* for multi-cpu machines */
+        options.clients++;
+        leave_critical_section(CRIT_CLIENTS);
     }
+    log(LOG_ERR, "INTERNAL ERROR: End of infinite loop");
 }
 
 #ifndef USE_WIN32
@@ -370,220 +418,6 @@ static int listen_local() { /* bind and listen on local interface */
     return ls;
 }
 
-int connect_local(u32 ip) { /* spawn local process */
-#ifdef USE_WIN32
-    log(LOG_ERR, "LOCAL MODE NOT SUPPORTED ON WIN32 PLATFORM");
-    return -1;
-#else
-    struct in_addr addr;
-    char text[STRLEN];
-    int fd[2];
-    unsigned long pid;
-
-    if (options.option & OPT_PTY) {
-        char tty[STRLEN];
-
-        if(pty_allocate(fd, fd+1, tty, STRLEN)) {
-            return -1;
-        }
-        log(LOG_DEBUG, "%s allocated", tty);
-    } else {
-        if(make_sockets(fd))
-            return -1;
-    }
-#ifdef USE_FORK
-    /* Each child has to take care of its own dead children */
-    signal(SIGCHLD, local_handler);
-#endif /* defined USE_FORK */
-    /* With USE_PTHREAD main thread does the work */
-    /* and SIGCHLD is blocked in other theads */
-    pid=(unsigned long)fork();
-    switch(pid) {
-    case -1:    /* error */
-        closesocket(fd[0]);
-        closesocket(fd[1]);
-        ioerror("fork");
-        return -1;
-    case  0:    /* child */
-        closesocket(fd[0]);
-        dup2(fd[1], 0);
-        dup2(fd[1], 1);
-        if (!options.foreground)
-            dup2(fd[1], 2);
-        closesocket(fd[1]);
-        if (ip) {
-            putenv("LD_PRELOAD=" libdir "/stunnel.so");
-            /* For Tru64 _RLD_LIST is used instead */
-            putenv("_RLD_LIST=" libdir "/stunnel.so:DEFAULT");
-            addr.s_addr = ip;
-            safecopy(text, "REMOTE_HOST=");
-            enter_critical_section(CRIT_NTOA); /* inet_ntoa is not mt-safe */
-            safeconcat(text, inet_ntoa(addr));
-            leave_critical_section(CRIT_NTOA);
-            putenv(text);
-        }
-        execvp(options.execname, options.execargs);
-        ioerror("execvp"); /* execv failed */
-        _exit(1);
-    }
-    /* parent */
-    log(LOG_INFO, "Local mode child started (PID=%lu)", pid);
-    closesocket(fd[1]);
-    return fd[0];
-#endif /* USE_WIN32 */
-}
-
-int connect_remote(u32 ip) { /* connect to remote host */
-    struct sockaddr_in addr;
-    int s; /* destination socket */
-    u32 *list; /* destination addresses list */
-
-    if((s=socket(AF_INET, SOCK_STREAM, 0))<0) {
-        sockerror("remote socket");
-        return -1;
-    }
-    memset(&addr, 0, sizeof(addr));
-    addr.sin_family=AF_INET;
-
-    if(ip) { /* transparent proxy */
-        addr.sin_addr.s_addr=ip;
-        addr.sin_port=htons(0);
-        if(bind(s, (struct sockaddr *)&addr, sizeof(addr))<0) {
-            sockerror("bind transparent");
-            return -1;
-        }
-    }
-
-    addr.sin_port=options.remoteport;
-
-    /* connect each host from the list*/
-    for(list=options.remotenames; *list!=-1; list++) {
-        addr.sin_addr.s_addr=*list;
-        enter_critical_section(CRIT_NTOA); /* inet_ntoa is not mt-safe */
-        log(LOG_DEBUG, "%s connecting %s:%d", options.servname,
-            inet_ntoa(addr.sin_addr), ntohs(addr.sin_port));
-        leave_critical_section(CRIT_NTOA);
-        if(!connect(s, (struct sockaddr *) &addr, sizeof(addr)))
-            return s; /* success */
-    }
-    sockerror("remote connect");
-    return -1;
-}
-
-int auth_user(struct sockaddr_in *addr) {
-    struct servent *s_ent;    /* structure for getservbyname */
-    struct sockaddr_in ident; /* IDENT socket name */
-    int s;                    /* IDENT socket descriptor */
-    char buff[STRLEN], name[STRLEN];
-    int ptr, len, retval;
-
-    if(!options.username)
-        return 0; /* -u option not specified */
-    if((s=socket(AF_INET, SOCK_STREAM, 0))<0) {
-        sockerror("socket (ident)");
-        return -1;
-    }
-    memcpy(&ident, addr, sizeof(ident));
-    s_ent=getservbyname("auth", "tcp");
-    if(!s_ent) {
-        log(LOG_WARNING, "Unknown service 'auth' - using default 113");
-        ident.sin_port=htons(113);
-    } else {
-        ident.sin_port=s_ent->s_port;
-    }
-    if(connect(s, (struct sockaddr *)&ident, sizeof(ident))<0) {
-        sockerror("connect (ident)");
-        closesocket(s);
-        return -1;
-    }
-#ifdef HAVE_SNPRINTF
-    len=snprintf(buff, STRLEN,
-        "%u , %u\r\n", ntohs(addr->sin_port), ntohs(options.localport));
-#else
-    len=sprintf(buff,
-        "%u , %u\r\n", ntohs(addr->sin_port), ntohs(options.localport));
-#endif
-    len=writesocket(s, buff, len);
-    if(len<0) {
-        sockerror("writesocket (ident)");
-        closesocket(s);
-        return -1;
-    }
-    ptr=0;
-    do {
-        len=readsocket(s, buff+ptr, STRLEN-ptr-1);
-        if(len<0) {
-            sockerror("readsocket (ident)");
-            closesocket(s);
-            return -1;
-        }
-        ptr+=len;
-    } while(len && ptr<STRLEN-1);
-    closesocket(s);
-    buff[ptr]='\0';
-    if(sscanf(buff, "%*[^:]: USERID :%*[^:]:%s", name)!=1) {
-        log(LOG_ERR, "Incorrect data from ident server");
-        return -1;
-    }
-    retval=strcmp(name, options.username) ? -1 : 0;
-    safestring(name);
-    log(LOG_INFO, "IDENT resolved remote user to %s", name);
-    return retval;
-}
-
-#ifndef USE_WIN32
-static int make_sockets(int fd[2]) { /* make pair of connected sockets */
-#ifdef INET_SOCKET_PAIR
-    struct sockaddr_in addr;
-    int addrlen;
-    int s; /* temporary socket awaiting for connection */
-
-    if((s=socket(AF_INET, SOCK_STREAM, 0))<0) {
-        sockerror("socket#1");
-        return -1;
-    }
-    if((fd[1]=socket(AF_INET, SOCK_STREAM, 0))<0) {
-        sockerror("socket#2");
-        return -1;
-    }
-    addrlen=sizeof(addr);
-    memset(&addr, 0, addrlen);
-    addr.sin_family=AF_INET;
-    addr.sin_addr.s_addr=htonl(INADDR_LOOPBACK);
-    addr.sin_port=0; /* dynamic port allocation */
-    if(bind(s, (struct sockaddr *)&addr, addrlen))
-        log(LOG_DEBUG, "bind#1: %s (%d)",
-            strerror(get_last_socket_error()), get_last_socket_error());
-    if(bind(fd[1], (struct sockaddr *)&addr, addrlen))
-        log(LOG_DEBUG, "bind#2: %s (%d)",
-            strerror(get_last_socket_error()), get_last_socket_error());
-    if(listen(s, 5)) {
-        sockerror("listen");
-        return -1;
-    }
-    if(getsockname(s, (struct sockaddr *)&addr, &addrlen)) {
-        sockerror("getsockname");
-        return -1;
-    }
-    if(connect(fd[1], (struct sockaddr *)&addr, addrlen)) {
-        sockerror("connect");
-        return -1;
-    }
-    if((fd[0]=accept(s, (struct sockaddr *)&addr, &addrlen))<0) {
-        sockerror("accept");
-        return -1;
-    }
-    closesocket(s); /* don't care about the result */
-#else
-    if(socketpair(AF_UNIX, SOCK_STREAM, 0, fd)) {
-        sockerror("socketpair");
-        return -1;
-    }
-#endif
-    return 0;
-}
-#endif
-
 int set_socket_options(int s, int type) {
     sock_opt *ptr;
     extern sock_opt sock_opts[];
@@ -615,27 +449,143 @@ int set_socket_options(int s, int type) {
     return 0; /* OK */
 }
 
-static void ioerror(char *txt) { /* Input/Output error handler */
-    int error;
-
-    error=get_last_error();
-    log(LOG_ERR, "%s: %s (%d)", txt, strerror(error), error);
+void ioerror(char *txt) { /* Input/Output error handler */
+    log_error(LOG_ERR, get_last_error(), txt);
 }
 
 void sockerror(char *txt) { /* Socket error handler */
-    int error;
+    log_error(LOG_ERR, get_last_socket_error(), txt);
+}
 
-    error=get_last_socket_error();
-    log(LOG_ERR, "%s: %s (%d)", txt, strerror(error), error);
+void log_error(int level, int error, char *txt) { /* Generic error logger */
+    log(level, "%s: %s (%d)", txt, my_strerror(error), error);
+}
+
+static char *my_strerror(int errnum) {
+    switch(errnum) {
+#ifdef USE_WIN32
+    case 10004:
+        return "Interrupted system call (WSAEINTR)";
+    case 10009:
+        return "Bad file number (WSAEBADF)";
+    case 10013:
+        return "Permission denied (WSAEACCES)";
+    case 10014:
+        return "Bad address (WSAEFAULT)";
+    case 10022:
+        return "Invalid argument (WSAEINVAL)";
+    case 10024:
+        return "Too many open files (WSAEMFILE)";
+    case 10035:
+        return "Operation would block (WSAEWOULDBLOCK)";
+    case 10036:
+        return "Operation now in progress (WSAEINPROGRESS)";
+    case 10037:
+        return "Operation already in progress (WSAEALREADY)";
+    case 10038:
+        return "Socket operation on non-socket (WSAENOTSOCK)";
+    case 10039:
+        return "Destination address required (WSAEDESTADDRREQ)";
+    case 10040:
+        return "Message too long (WSAEMSGSIZE)";
+    case 10041:
+        return "Protocol wrong type for socket (WSAEPROTOTYPE)";
+    case 10042:
+        return "Bad protocol option (WSAENOPROTOOPT)";
+    case 10043:
+        return "Protocol not supported (WSAEPROTONOSUPPORT)";
+    case 10044:
+        return "Socket type not supported (WSAESOCKTNOSUPPORT)";
+    case 10045:
+        return "Operation not supported on socket (WSAEOPNOTSUPP)";
+    case 10046:
+        return "Protocol family not supported (WSAEPFNOSUPPORT)";
+    case 10047:
+        return "Address family not supported by protocol family (WSAEAFNOSUPPORT)";
+    case 10048:
+        return "Address already in use (WSAEADDRINUSE)";
+    case 10049:
+        return "Can't assign requested address (WSAEADDRNOTAVAIL)";
+    case 10050:
+        return "Network is down (WSAENETDOWN)";
+    case 10051:
+        return "Network is unreachable (WSAENETUNREACH)";
+    case 10052:
+        return "Net dropped connection or reset (WSAENETRESET)";
+    case 10053:
+        return "Software caused connection abort (WSAECONNABORTED)";
+    case 10054:
+        return "Connection reset by peer (WSAECONNRESET)";
+    case 10055:
+        return "No buffer space available (WSAENOBUFS)";
+    case 10056:
+        return "Socket is already connected (WSAEISCONN)";
+    case 10057:
+        return "Socket is not connected (WSAENOTCONN)";
+    case 10058:
+        return "Can't send after socket shutdown (WSAESHUTDOWN)";
+    case 10059:
+        return "Too many references, can't splice (WSAETOOMANYREFS)";
+    case 10060:
+        return "Connection timed out (WSAETIMEDOUT)";
+    case 10061:
+        return "Connection refused (WSAECONNREFUSED)";
+    case 10062:
+        return "Too many levels of symbolic links (WSAELOOP)";
+    case 10063:
+        return "File name too long (WSAENAMETOOLONG)";
+    case 10064:
+        return "Host is down (WSAEHOSTDOWN)";
+    case 10065:
+        return "No Route to Host (WSAEHOSTUNREACH)";
+    case 10066:
+        return "Directory not empty (WSAENOTEMPTY)";
+    case 10067:
+        return "Too many processes (WSAEPROCLIM)";
+    case 10068:
+        return "Too many users (WSAEUSERS)";
+    case 10069:
+        return "Disc Quota Exceeded (WSAEDQUOT)";
+    case 10070:
+        return "Stale NFS file handle (WSAESTALE)";
+    case 10091:
+        return "Network SubSystem is unavailable (WSASYSNOTREADY)";
+    case 10092:
+        return "WINSOCK DLL Version out of range (WSAVERNOTSUPPORTED)";
+    case 10093:
+        return "Successful WSASTARTUP not yet performed (WSANOTINITIALISED)";
+    case 10071:
+        return "Too many levels of remote in path (WSAEREMOTE)";
+    case 11001:
+        return "Host not found (WSAHOST_NOT_FOUND)";
+    case 11002:
+        return "Non-Authoritative Host not found (WSATRY_AGAIN)";
+    case 11003:
+        return "Non-Recoverable errors: FORMERR, REFUSED, NOTIMP (WSANO_RECOVERY)";
+    case 11004:
+        return "Valid name, no data record of requested type (WSANO_DATA)";
+#if 0
+    case 11004: /* typically, only WSANO_DATA is reported */
+        return "No address, look for MX record (WSANO_ADDRESS)";
+#endif
+#endif /* defined USE_WIN32 */
+    default:
+        return strerror(errnum);
+    }
 }
 
 #ifdef USE_FORK
 static void sigchld_handler(int sig) { /* Dead children detected */
     int pid, status;
 
-#if defined(HAVE_WAITPID)
-    while((pid=waitpid(-1, &status, WNOHANG))>0) {
+#ifdef HAVE_WAIT_FOR_PID
+    while((pid=wait_for_pid(-1, &status, WNOHANG))>0) {
         options.clients--; /* One client less */
+#else
+    if((pid=wait(&status))>0) {
+        options.clients--; /* One client less */
+#endif
+#ifdef WIFSIGNALED
         if(WIFSIGNALED(status)) {
             log(LOG_DEBUG, "%s[%d] terminated on signal %d (%d left)",
                 options.servname, pid, WTERMSIG(status), options.clients);
@@ -645,10 +595,9 @@ static void sigchld_handler(int sig) { /* Dead children detected */
         }
     }
 #else
-    pid=wait(&status);
-    options.clients--; /* One client less */
-    log(LOG_DEBUG, "%s[%d] finished with code %d (%d left)",
-        options.servname, pid, status, options.clients);
+        log(LOG_DEBUG, "%s[%d] finished with code %d (%d left)",
+            options.servname, pid, status, options.clients);
+    }
 #endif
     signal(SIGCHLD, sigchld_handler);
 }
@@ -656,12 +605,27 @@ static void sigchld_handler(int sig) { /* Dead children detected */
 
 #ifndef USE_WIN32
 
-static void local_handler(int sig) { /* sigchld handler for -l processes */
+void local_handler(int sig) { /* Dead of local (-l) process detected */
     int pid, status;
 
-    pid=wait(&status);
-    log(LOG_DEBUG, "Local process %s[%d] finished with code %d",
-        options.servname, pid, status);
+#ifdef HAVE_WAIT_FOR_PID
+    while((pid=wait_for_pid(-1, &status, WNOHANG))>0) {
+#else
+    if((pid=wait(&status))>0) {
+#endif
+#ifdef WIFSIGNALED
+        if(WIFSIGNALED(status)) {
+            log(LOG_DEBUG, "Local process %s (PID=%lu) terminated on signal %d",
+                options.servname, pid, WTERMSIG(status));
+        } else {
+            log(LOG_DEBUG, "Local process %s (PID=%lu) finished with code %d",
+                options.servname, pid, WEXITSTATUS(status));
+        }
+#else
+        log(LOG_DEBUG, "Local process %s (PID=%lu) finished with status %d",
+            options.servname, pid, status);
+#endif
+    }
     signal(SIGCHLD, local_handler);
 }
 
@@ -670,6 +634,52 @@ static void signal_handler(int sig) { /* Signal handler */
     exit(3);
 }
 
+#else /* !defined USE_win32 */
+
+static BOOL CtrlHandler(DWORD fdwCtrlType) {
+    switch(fdwCtrlType) {
+    case CTRL_C_EVENT: /* Handle the CTRL+C signal */
+        log(LOG_NOTICE, "Process exit: CTRL+C");
+        break;
+    case CTRL_CLOSE_EVENT: /* CTRL+CLOSE: confirm that the user wants to exit */
+        log(LOG_NOTICE, "Process exit: window closed");
+        break;
+    case CTRL_BREAK_EVENT:
+        log(LOG_NOTICE, "Process exit: CTRL+BREAK");
+        break;
+    case CTRL_LOGOFF_EVENT:
+        log(LOG_NOTICE, "Process exit: user logoff");
+        break;
+    case CTRL_SHUTDOWN_EVENT:
+        log(LOG_NOTICE, "Process exit: system shutdown");
+        break;
+    default:
+        return FALSE;
+    }
+    ExitProcess(0);
+}
+
 #endif /* !defined USE_WIN32 */
 
+char *stunnel_info() {
+    static char retval[STRLEN];
+
+    safecopy(retval, "stunnel " VERSION " on " HOST);
+#ifdef USE_PTHREAD
+    safeconcat(retval, " PTHREAD");
+#endif
+#ifdef USE_WIN32
+    safeconcat(retval, " WIN32");
+#endif
+#ifdef USE_FORK
+    safeconcat(retval, " FORK");
+#endif
+#ifdef USE_LIBWRAP
+    safeconcat(retval, "+LIBWRAP");
+#endif
+    safeconcat(retval, " with ");
+    safeconcat(retval, SSLeay_version(SSLEAY_VERSION));
+    return retval;
+}
+
 /* End of stunnel.c */
